#include <stdio.h> 
#include <stdint.h> 
#include <string.h> 
#include <stdlib.h> 
#include <sys/types.h> 
#include <errno.h> 
#include "qumap.h"

typedef struct
{
  union { struct { void* child[2];           } inte;  
          struct { void *data; uint64_t key; } leaf; } d;
  unsigned internal:1, bit:6;
} qumap_node_t;

#define Q_KEY(X)    ((X)->d.leaf.key)
#define Q_DATA(X)   ((X)->d.leaf.data)
#define Q_CHLD(X,Y) ((X)->d.inte.child[Y])


static inline int qumap_dir(uint64_t key,qumap_node_t *node)
{
  return (key >> node->bit)&1;
}

void* qumap_find(qumap_t *map,uint64_t key)
{
  qumap_node_t *node;

  node= map->root;
  if (!node) return NULL;

  while(node->internal) node= Q_CHLD(node, qumap_dir(key,node));

  if (Q_KEY(node)==key) return Q_DATA(node);
                          else return NULL;
}



static inline qumap_node_t* qumap_leaf_node(uint64_t key, void *data)
{
  qumap_node_t *N;
  N= malloc(sizeof(*N));
  N->internal= 0;
  Q_KEY(N)= key;
  Q_DATA(N)= data; 
  return N;
}

static inline qumap_node_t* qumap_internal_node(uint32_t bit)
{
  qumap_node_t *N;
  N= malloc(sizeof(*N));
  N->internal= 1;
  N->bit= bit;
  return N;
}

static uint8_t qumap_crit_bit(uint64_t X, uint64_t Y)
{
  uint32_t i,j;
  static const uint64_t B[]=
  {
    0x5555555555555555ll,
    0x3333333333333333ll,
    0x0f0f0f0f0f0f0f0fll,
    0x00ff00ff00ff00ffll,
    0x0000ffff0000ffffll,
    0x00000000ffffffffll
  };
  uint64_t A;
  A= X ^ Y;
  
  for(i=1;i<64;i<<=1) A |= (A<<i);
  A = ~A;
  for(j=0,i=1;i<64;i<<=1,j++) A= (A&B[j]) + ((A>>i)&B[j]);
  return A;
}

void* qumap_insert(qumap_t* map, uint64_t key, void *data)
{
  qumap_node_t *node_elt;

  if (!map->root)
  {
     map->root= qumap_leaf_node(key,data);
     return 0;
  }

  node_elt= map->root;
  while(node_elt->internal)
    node_elt= Q_CHLD(node_elt, qumap_dir(key,node_elt));

  
  uint64_t elt;
  int dir_elt;
  uint8_t bit;
  qumap_node_t *node_key;

  elt= Q_KEY(node_elt);

  if (elt==key) return Q_DATA(node_elt);

  bit= qumap_crit_bit(elt,key);

  dir_elt= (elt>>bit)&1;

  node_key= qumap_internal_node(bit);
  Q_CHLD(node_key,1-dir_elt)= qumap_leaf_node(key,data);

  qumap_node_t *parent;
  qumap_node_t *node;

  parent= 0;
  node= map->root;

  while(node->internal)
  {
     if (node->bit>bit) break;
     parent= node;
     node= Q_CHLD(node, qumap_dir(key,node));
  }

  Q_CHLD(node_key,dir_elt)= node;
  if (parent) { if (Q_CHLD(parent,0)==node)  Q_CHLD(parent,0)= node_key;
                                       else  Q_CHLD(parent,1)= node_key; }
  else map->root= node_key;
  return 0;
}

void *qumap_remove(qumap_t* map, uint64_t key)
{
  void *retv;

  if (!map->root) return NULL;

   
  qumap_node_t *node;
  qumap_node_t *sibling;
  qumap_node_t *parent;
  qumap_node_t *gparent;

  node= map->root;
  gparent= 0;
  sibling= 0;
  parent= 0;

  while(node->internal)
  {
    int dir;
    dir= qumap_dir(key, node);
    gparent= parent;
    parent= node;
    sibling= Q_CHLD(node, 1-dir);
    node= Q_CHLD(node, dir);
  }

  if (Q_KEY(node)!=key) return NULL;
  retv= Q_DATA(node);
  free(node);

  if (!gparent)
  {
     map->root= sibling;
     if (parent) free(parent);
     return retv;
  }

  if (Q_CHLD(gparent,0)==parent) Q_CHLD(gparent,0)= sibling;
                            else Q_CHLD(gparent,1)= sibling;
  free(parent);

  return retv;
}

void *qumap_destroy(qumap_t *map)
{
  qumap_node_t *node;
  qumap_node_t *parent;
  qumap_node_t *gparent;
  qumap_node_t *sibling;
  void *retv;

  if (!map->root) return 0;

  parent= 0;
  gparent= 0;
  sibling= 0;
  node= map->root;
  while(node->internal)
  {
    gparent= parent;
    parent= node;
    sibling= Q_CHLD(node, 1);
    node= Q_CHLD(node,0);
  }
  retv= Q_DATA(node);
  free(node);
  if (!gparent) 
  {
    map->root= sibling;
    if (parent) free(parent);
    return retv;
  }
  if (Q_CHLD(gparent,0)==parent)
    Q_CHLD(gparent,0)= sibling;
  else
    Q_CHLD(gparent,1)= sibling;

  free(parent);

  return retv;
}

qumap_t  *qumap_new()
{
  qumap_t *s;
  s= malloc(sizeof(*s));
  s->root= 0;
  return s;
}
